﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// This timestamp runs on its own and it
    /// uses very little CPU usage to keep track of time. This is a very
    /// useful system. This timestamp is in elapsed miliseconds.
    /// </summary>
    internal class GlobalTimeStamp : IDisposable
    {
        /// <summary>
        /// This is the instance of the time stamp which is used
        /// to auto start this.
        /// </summary>
        public static GlobalTimeStamp instance = new GlobalTimeStamp();

        /// <summary>
        /// The Current Timestamp, elapsed time since the begining in Miliseconds.
        /// </summary>
        public static ulong Current
        {
            get
            {
                return instance._Current;
            }
        }

        /// <summary>
        /// The StopWatch that the TimeStamp uses
        /// </summary>
        private Stopwatch Watch = null;

        /// <summary>
        /// Whether or not this Timestamp Is Running.
        /// </summary>
        private bool Running = false;

        /// <summary>
        /// The Refresh Thread for this GlobalTimeStamp.
        /// </summary>
        private Thread RefreshThread = null;

        /// <summary>
        /// The Current Timestamp, elapsed time since the begining in Miliseconds.
        /// </summary>
        public ulong _Current = 777; //Initialize this value so we have something.

        /// <summary>
        /// A new instance of the GlobalTimeStamp
        /// </summary>
        public GlobalTimeStamp()
        {
            //Start the watch.
            this.Watch = new Stopwatch();
            this.Watch.Start();

            //Setup the Refresh Thread
            this.Running = true;
            this.RefreshThread = new Thread(Refresh);
            this.RefreshThread.IsBackground = true;
            this.RefreshThread.Start();
        }

        /// <summary>
        /// The method that handles Refreshing the TimeStamp.
        /// </summary>
        private void Refresh()
        {
            while (this.Running)
            {
                //Get the elapsed miliseconds from the watch
                this._Current = (ulong)this.Watch.ElapsedMilliseconds;

                //Now sleep 1MS
                Thread.Sleep(1);
            }
        }

        /// <summary>
        /// Returns the Difference between two TimeStamps in miliseconds.
        /// Note:
        ///     We needed this function because sometimes we
        ///     store a timestamp of when an event occured but when
        ///     using multi threading if this is greater than the last
        ///     time we got the current time, then we need to select the bigger one.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns>The difference in miliseconds.</returns>
        public static ulong GetDifference(ulong a, ulong b)
        {
            if (a > b)
                return (a - b);

            return (b - a);
        }

        /// <summary>
        /// Returns the Difference in miliseconds between the Current Time and the passed
        /// Time stamp.
        /// Note:
        ///     We needed this function because sometimes we
        ///     store a timestamp of when an event occured but when
        ///     using multi threading if this is greater than the last
        ///     time we got the current time, then we need to select the bigger one.
        /// </summary>
        /// <param name="b"></param>
        /// <returns>The difference in miliseconds.</returns>
        public static ulong GetDifference(ulong b)
        {
            //Get the current timestamp
            ulong current = instance._Current;
            //Compare it.
            if (current > b)
                return (current - b);

            return (b - current);
        }

        /// <summary>
        /// Disposes this GlobalTimeStamp if its Running.
        /// </summary>
        public void Dispose()
        {
            try
            {
                if (this.Running)
                {
                    this.Running = false;

                    this.Watch.Stop();

                    this.RefreshThread = null;
                }
            }
            catch { }
        }

        /// <summary>
        /// Called when this GlobalTimeStamp is going to be destroyed.
        /// </summary>
        ~GlobalTimeStamp()
        {
            try
            {
                if (this.Running)
                {
                    this.Running = false;

                    this.Watch.Stop();

                    this.RefreshThread = null;
                }
            }
            catch { }
        }
    }
}
