﻿using MySql.Data.MySqlClient;
using System;
using System.Linq;
using System.Reflection;
using System.Text;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// A set of extensions to help work with MySqlCommands.
    /// </summary>
    public static class MySqlCommandExtensions
    {
        /// <summary>
        /// Setups this MySqlCommand to insert a row in the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Insert<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get all the fields.
            var fields = rowType.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            //Start building the query.
            var builder = new StringBuilder();

            //Write the insert section.
            builder.Append($"INSERT INTO `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Write the set values section.
            builder.Append("SET ");

            bool seperate = false;
            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;

                //Skip any columns that are auto incremented.
                if (column.AutoIncrement)
                    continue;

                if (seperate)
                    builder.Append(", ");

                builder.Append($"`{(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)}` = @{i}");

                command.Parameters.AddWithValue($"@{i}", (column.Converter == null ? fields[i].GetValue(row) : column.Converter.CreateInstance<MySqlColumnConverter>().ConvertTo(fields[i].GetValue(row))));

                seperate = true;
            }

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to update a row in the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Update<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get all the fields.
            var fields = rowType.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            //Start building the query.
            var builder = new StringBuilder();

            //Write the update section
            builder.Append($"UPDATE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Write the set values section
            builder.Append("SET ");

            FieldInfo primaryField = null;
            MySqlColumn primaryColumn = null;
            bool seperate = false;
            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;

                //If this column is the primary key, grab it.
                if (column.PrimaryKey)
                {
                    primaryField = fields[i];
                    primaryColumn = column;
                }

                //Skip any columns that are auto incremented.
                if (column.AutoIncrement)
                    continue;

                if (seperate)
                    builder.Append(", ");

                builder.Append($"`{(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)}` = @{i}");

                command.Parameters.AddWithValue($"@{i}", (column.Converter == null ? fields[i].GetValue(row) : column.Converter.CreateInstance<MySqlColumnConverter>().ConvertTo(fields[i].GetValue(row))));

                seperate = true;
            }

            //Make sure we have a primary key.
            if (primaryField == null)
                throw new Exception("Given row does not contain primary key.");

            //Write the where clause
            builder.Append($" WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1");

            command.Parameters.AddWithValue("@id", primaryField.GetValue(row));

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to get a row from the db by it's id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        /// <exception cref="Exception"></exception>
        public static void Get<T>(this MySqlCommand command, ulong id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get the id field
            if (!rowType.GetMySqlPrimaryKey(out FieldInfo primaryField, out MySqlColumn primaryColumn))
                throw new Exception("Failed to find the primary key for the given row.");

            //Set the command text.
            command.CommandText = $"SELECT * FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1";
            
            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);
        }

        /// <summary>
        /// Setups this MySqlCommand to get a row from the db by a string id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        public static void Get<T>(this MySqlCommand command, string id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get the id field
            if (!rowType.GetMySqlPrimaryKey(out FieldInfo primaryField, out MySqlColumn primaryColumn))
                throw new Exception("Failed to find the primary key for the given row.");

            //Set the command text.
            command.CommandText = $"SELECT * FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1";

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);
        }

        /// <summary>
        /// Setups this MySqlCommand to get all the rows of a given type from the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void GetAll<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Set the command text.
            command.CommandText = $"SELECT * FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`";
        }

        /// <summary>
        /// Setups this MySqlCommand to see if a row exists by an id from the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        public static void Exists<T>(this MySqlCommand command, ulong id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get the id field
            if (!rowType.GetMySqlPrimaryKey(out FieldInfo primaryField, out MySqlColumn primaryColumn))
                throw new Exception("Failed to find the primary key for the given row.");

            //Set the command text.
            command.CommandText = $"SELECT EXISTS(SELECT 1 FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1)";

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);
        }

        /// <summary>
        /// Setups this MySqlCommand to see if a row exists by an id from the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        public static void Exists<T>(this MySqlCommand command, string id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Get the id field
            if (!rowType.GetMySqlPrimaryKey(out FieldInfo primaryField, out MySqlColumn primaryColumn))
                throw new Exception("Failed to find the primary key for the given row.");

            //Set the command text.
            command.CommandText = $"SELECT EXISTS(SELECT 1 FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1)";

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a row from the db by it's id from an instance of the row.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Delete<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Get the primary field and column
            if (!row.GetMySqlPrimaryKey(out FieldInfo primaryField, out MySqlColumn primaryColumn))
                throw new Exception("Failed to find primary key for given row.");

            //Write the where clause
            builder.Append($"WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1");

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", primaryField.GetValue(row));

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a row from the db by it's id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        public static void Delete<T>(this MySqlCommand command, ulong id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Get the primary key field and column.
            if (!rowType.GetMySqlPrimaryKey(out FieldInfo primaryField, out MySqlColumn primaryColumn))
                throw new Exception("Failed to find primary key for the given row.");

            //Write the where clause
            builder.Append($"WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1");

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a row from the db by it's string id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        public static void Delete<T>(this MySqlCommand command, string id)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` ");

            //Get the primary key field and column.
            if (!rowType.GetMySqlPrimaryKey(out FieldInfo primaryField, out MySqlColumn primaryColumn))
                throw new Exception("Failed to find primary key for the given row.");

            //Write the where clause
            builder.Append($"WHERE `{(string.IsNullOrWhiteSpace(primaryColumn.Name) ? primaryField.Name : primaryColumn.Name)}`=@id LIMIT 1");

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete all rows of a given type from the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void DeleteAll<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`");

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to create a table from a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void CreateTable<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Create a new instance of the row
            var row = rowType.CreateInstance<T>();

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"CREATE TABLE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}` (");

            //Write all the columns
            var fields = rowType.GetFields();

            bool separate = false;

            if (fields != null)
            {
                for (int i = 0; i < fields.Length; i++)
                {
                    var field = fields[i];

                    var column = field.GetCustomAttribute<MySqlColumn>();

                    if (column != null)
                    {
                        if (separate)
                            builder.Append(", ");

                        var columnName = string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name;

                        //Write the column name
                        builder.Append($"{columnName} ");

                        //Write the column data type
                        builder.Append($"{column.GetMySqlColumnType(fields[i])} ");

                        //Determine if this column is nullable
                        bool nullable = column.IsNullable(field, row);

                        //Write the column null information
                        builder.Append($"{(nullable ? "NULL" : "NOT NULL")} ");

                        //Write the column default value if needed. (We can't do this if the column is auto increment)
                        if (!column.AutoIncrement)
                        {
                            var defaultValue = column.DefaultValue;

                            //If we don't have a default value, attempt to get it from the field.
                            if (defaultValue == null)
                                defaultValue = field.GetValue(row);

                            //If we have a converter, run the converter to get the true default value.
                            if (column.Converter != null)
                                defaultValue = column.Converter.CreateInstance<MySqlColumnConverter>().ConvertTo(defaultValue);

                            //If we have a default value, set it on the column.
                            if (defaultValue != null)
                            {
                                builder.Append($"DEFAULT @{i} ");

                                command.Parameters.AddWithValue($"@{i}", defaultValue);
                            }
                        }

                        //Write the column auto increment information.
                        builder.Append($"{(column.AutoIncrement ? "AUTO_INCREMENT" : "")}");

                        //If the column is the primary key, output that constraint. 
                        if (column.PrimaryKey)
                            builder.Append($", PRIMARY KEY ({columnName})");
                        //If the column is unique, output that constraint, but you can't be the primary key and be unique.
                        else if (column.Unique)
                            builder.Append($", UNIQUE ({columnName})");

                        separate = true;
                    }
                }
            }

            //Add any indexes
            var indexes = rowType.GetCustomAttributes<MySqlRowIndex>().ToList();
            if (indexes != null)
            {
                for (int i = 0; i < indexes.Count; i++)
                {
                    var index = indexes[i];

                    if (index.Columns == null || index.Columns.Length == 0)
                        continue;

                    if (separate)
                        builder.Append(", ");

                    if (string.IsNullOrWhiteSpace(index.Name))
                        builder.Append($"INDEX IDX_{index.Columns.Aggregate((curr, next) => curr == null ? next : curr + "_" + next)} ({index.Columns.Aggregate((curr, next) => curr == null ? next : curr + ", " + next)})");
                    else
                        builder.Append($"INDEX {index.Name} ({index.Columns.Aggregate((curr, next) => curr == null ? next : curr + ", " + next)})");

                    separate = true;
                }
            }

            builder.Append(")");

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to check if a table exists for a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void TableExists<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"SHOW TABLES LIKE '{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}'");

            //Set the command text
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to empty the table for a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void EmptyTable<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"TRUNCATE TABLE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`");

            //Set the command text
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a table for a given row type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void DeleteTable<T>(this MySqlCommand command)
        {
            //Get the type of T
            var rowType = typeof(T);

            //Get the row information.
            var rowAttribute = rowType.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DROP TABLE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? rowType.Name : rowAttribute.Name)}`");

            //Set the command text
            command.CommandText = builder.ToString();
        }
    }
}
